/*
    btimer.cpp

    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997-2003  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#ifdef linux
#include <signal.h>
#include <sys/time.h>
#endif
#ifdef WIN32
#include <process.h>
#endif

#include "btimer.h"

#ifdef WIN32
#ifndef _TTHREADPROC_DEFINED_
typedef  unsigned int ( __stdcall *tThreadProc )( void * );
#define _TTHREADPROC_DEFINED_
#endif
#endif

BTimer *BTimer::instance = NULL;

BTimer::BTimer() : dueTime(0), periodic(false),
                   timerProc(NULL), timerParam(NULL)
#ifdef WIN32
		   , timerObject(NULL), threadId(0)
#endif
{
	instance = this;
	Init();
}

BTimer::~BTimer()
{
	Stop();
	UnInit();
	instance  = NULL;
	timerProc = NULL;
}

BTimer *BTimer::Instance()
{
	if (instance == NULL)
		instance = new BTimer;
	return instance;
}

void BTimer::finalize()
{
	delete instance;
	instance = NULL;
}

bool BTimer::Start(bool x_periodic, uint64_t x_dueTime)
{
	dueTime  = x_dueTime;
	periodic = x_periodic;
	return StartTimer();
}

void BTimer::SetTimerProc(BTimerProc x_timerProc, void *x_p /* = NULL*/)	
{
	timerParam = x_p;
	timerProc  = x_timerProc;
}

///////////////////////////////////////////////////////////////
// BTimer linux specific implementation
///////////////////////////////////////////////////////////////
#ifdef linux
void BTimer::Init()
{
}

void BTimer::UnInit()
{
}

bool BTimer::Stop(void)
{
	bool ret = true;
        struct itimerval timerValue;

	periodic   = false;
	timerProc  = NULL;
	timerParam = NULL;

        timerValue.it_value.tv_usec    = 0;
        timerValue.it_value.tv_sec     = 0;
        timerValue.it_interval.tv_usec = 0;
        timerValue.it_interval.tv_sec  = 0;

#ifdef SIGALRM
        (void)signal(SIGALRM, NULL);
#endif
        ret = (setitimer(ITIMER_REAL, &timerValue, NULL) == 0);
	return ret;
}

void BTimer::Suspend(void)
{
	pause();
}

bool BTimer::StartTimer(void)
{
	bool ret = false;
#ifdef SIGALRM
        struct itimerval timerValue;

        timerValue.it_value.tv_usec    = dueTime % 1000000;
        timerValue.it_value.tv_sec     = dueTime / 1000000;
        timerValue.it_interval.tv_usec = dueTime % 1000000;
        timerValue.it_interval.tv_sec  = dueTime / 1000000;

        (void)signal(SIGALRM, UnixTimerCallback);
        ret = (setitimer(ITIMER_REAL, &timerValue, NULL) == 0);
#else
#error "No SIGALRM available. Can't use timer" 
#endif
	return ret;
}

RETSIGTYPE BTimer::UnixTimerCallback(int i)
{
	if (instance != NULL)
		instance->TimerElapsed();
}

void BTimer::TimerElapsed()
{
	if (timerProc != NULL)
		(*timerProc)(timerParam);
	if (periodic)
		StartTimer();
}
#endif

///////////////////////////////////////////////////////////////
// BTimer Win32 specific implementation
///////////////////////////////////////////////////////////////

#ifdef WIN32
void BTimer::Init()
{
	UnInit();
	timerObject = CreateWaitableTimer(NULL, true, NULL);
	if (timerObject == NULL)
		DEBUGPRINT1("CreateWaitableTimer failed (%d)\n", GetLastError())
}

void BTimer::UnInit()
{
	if (timerObject != NULL)
	{
		bool ret = (CancelWaitableTimer(timerObject) != 0);
		if (!ret)
			DEBUGPRINT1("CancelWaitableTimer failed (%d)\n", GetLastError());
		CloseHandle(timerObject);
		timerObject = NULL;
	}
}

bool BTimer::Stop(void)
{
	if (instance == NULL)
		return true;

	bool ret = false;

	periodic   = false;
	timerProc  = NULL;
	timerParam = NULL;
	if (timerObject != NULL)
	{
		ret = (CancelWaitableTimer(timerObject) != 0);
		if (!ret)
			DEBUGPRINT1("CancelWaitableTimer failed (%d)\n", GetLastError());
		CloseHandle(timerObject);
		timerObject = NULL;
	}
	return ret;
}

void BTimer::Suspend(void)
{
	DWORD ret;

	if (periodic && timerObject != NULL)
	{
		ret = WaitForSingleObject(timerObject, INFINITE);
		if (ret == WAIT_FAILED)
			DEBUGPRINT1("Suspend>>WaitForSingleObject failed (%d)\n", GetLastError());

	}
}

bool BTimer::StartTimer()
{
	unsigned long thread = _beginthreadex(NULL, 1024,
		(tThreadProc)BTimer::StartTimerImp, this, 0, &threadId);
	return (thread != 0);
}

unsigned int BTimer::StartTimerImp(BTimer *p)
{
	bool success = true;
	if (p == NULL || p->timerObject == NULL)
		return 1;

	LARGE_INTEGER aDueTime;
	DWORD ret;

	// increase Thread priority to avoid delays if timer elapses
	HANDLE hThread = (HANDLE)GetCurrentThread();
	SetThreadPriority(hThread, GetThreadPriority(hThread)+1);

	aDueTime.QuadPart = -((LONGLONG)p->dueTime * 10);

	do {
		success = (SetWaitableTimer(p->timerObject,
			&aDueTime, 0, NULL, NULL, false) != 0);	
		if (!success)
		{
			DEBUGPRINT1("SetWaitableTimer failed (%d)\n", GetLastError());
		}

		ret = WaitForSingleObject(p->timerObject, INFINITE);
		if (ret == WAIT_FAILED)
			DEBUGPRINT1("StartTimerImp>>WaitForSingleObject failed (%d)\n", GetLastError());
		if (ret == WAIT_OBJECT_0 && p->timerProc != NULL)
			(*p->timerProc)(p->timerParam);
	} while (p->periodic && ret == WAIT_OBJECT_0);
	DEBUGPRINT("Return from StartTimerImp\n");
	// end thread
	return 0;
}
#endif
