/*
    mc6809.h

    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997-2003  W. Schwotzer

    This file is based on usim-0.91 which is
    Copyright (C) 1994 by R. B. Bellis
*/



#ifndef __mc6809_h__
#define __mc6809_h__

#include <stdio.h>
#include "misc1.h"
#include "typedefs.h"
#include "memory.h"
#include "cpustate.h"


#define DO_NMI			0x01
#define DO_FIRQ			0x02
#define DO_IRQ			0x04
#define DO_INVALID		0x08
#define DO_BREAKPOINT		0x10
#define DO_SINGLESTEP		0x20
#define DO_SINGLESTEPFINISHED	0x40
#define DO_SYNCEXEC		0x80
#define DO_TIMER		0x100
#define DO_SET_STATUS		0x200
#define DO_FREQ_CONTROL		0x400
#define DO_GO_BACK		0x800
#define DO_LOG			0x1000
#define DO_CWAI			0x2000
#define DO_SYNC			0x4000
#define IGNORE_BP               0x8000

#define SINGLESTEP_OVER		0x54
#define SINGLESTEP_INTO		0x55
#define START_RUNNING		0x56
#define CONTINUE_RUNNING	0x57


#define SYNC_INSTR		1
#define CWAI_INSTR		2


// uncomment this for debug output of cpu
//#define	DEBUG_FILE	"cpu.txt"


struct s_cpu_logfile {
	unsigned int minAddr;
	unsigned int maxAddr;
	unsigned int startAddr;
	unsigned int stopAddr;
	char logFileName[PATH_MAX];
};

class BTime;
class Da6809;


typedef unsigned long int t_cycles; // must be unsigned long because of sprintf

class Mc6809 : public IoDevice {

// Processor registers
protected:

	Byte			indexed_cycles[256];	// add. cycles for
							// indexed addr.
	Byte			psh_pul_cycles[256];	// add. cycles for psh
							// and pull-instr. 
	Byte			nmi_armed;		// for handling
							// interrupts
	Word			events;			// event status flags
#ifdef FASTFLEX
	Word			ipcreg, iureg, isreg, ixreg, iyreg;
	Byte			iareg, ibreg, iccreg, idpreg;
	Word			eaddr;
	Byte			ireg;
	Byte			iflag;
	Byte			tb;
	Word			tw;
	Byte			k;
	Byte			*pMem;		// needed for memory access
#else
	Word			pc;
	Word			u, s;		// Stack pointers
	Word			x, y;		// Index registers
	union {
		Word			d;	// Combined accumulator
		struct {
#ifdef WORDS_BIGENDIAN
			Byte		a;	// Accumulator a
			Byte		b;	// Accumulator b
#else
			Byte		b;	// Accumulator b
			Byte		a;	// Accumulator a
#endif
		} byte;
	} acc;
	union {
		Word			dp16;
		struct {
#ifdef WORDS_BIGENDIAN
			Byte		dph;
			Byte		dpl;
#else
			Byte		dpl;
			Byte		dph;
#endif
		} byte;
	} dpreg;
	Byte&			a;
	Byte&			b;
	Word&			d;
	Byte&			dp;
	union {
		Byte			all;	// Condition code register
		struct {
#ifdef BITFIELDS_LSB_FIRST
			bool		c : 1;	// Carry
			bool		v : 1;	// Overflow
			bool		z : 1;	// Zero
			bool		n : 1;	// Negative
			bool		i : 1;	// IRQ disable
			bool		h : 1;	// Half carry
			bool		f : 1;	// FIRQ disable
			bool		e : 1;	// Entire
#else
			bool		e : 1;	// Entire
			bool		f : 1;	// FIRQ disable
			bool		h : 1;	// Half carry
			bool		i : 1;	// IRQ disable
			bool		n : 1;	// Negative
			bool		z : 1;	// Zero
			bool		v : 1;	// Overflow
			bool		c : 1;	// Carry
#endif
		} bit;
	} cc;
#endif // FASTFLEX

protected:

	Da6809*		disassembler;

// funcitons for instruction execution:

private:

	void			init(void);
	void			illegal();
#ifndef FASTFLEX
	inline Word		fetch_ea_ext(void);
	inline Word		fetch_ea_dir(void);
	inline Word		fetch_ea_idx(t_cycles *c);
	inline Byte		fetch_imm_08(void);
	inline Byte		fetch_ext_08(void);
	inline Byte		fetch_dir_08(void);
	inline Byte		fetch_idx_08(t_cycles *c);
	inline Word		fetch_imm_16(void);
	inline Word		fetch_ext_16(void);
	inline Word		fetch_dir_16(void);
	inline Word		fetch_idx_16(t_cycles *c);

	inline Word		do_effective_address(Byte);

	inline void		abx();
	inline void		and_(Byte &reg, Byte operand);
	inline void		andcc(Byte operand);
	inline void		asr(Byte &reg);
	inline void		asr(Word addr);
	inline void		bcc();
	inline t_cycles		lbcc();
	inline void		bcs();
	inline t_cycles		lbcs();
	inline void		beq();
	inline t_cycles		lbeq();
	inline void		bge();
	inline t_cycles		lbge();
	inline void		bgt();
	inline t_cycles		lbgt();
	inline void		bhi();
	inline t_cycles		lbhi();
	inline void		ble();
	inline t_cycles		lble();
	inline void		bls();
	inline t_cycles		lbls();
	inline void		blt();
	inline t_cycles		lblt();
	inline void		bmi();
	inline t_cycles		lbmi();
	inline void		bne();
	inline t_cycles		lbne();
	inline void		bpl();
	inline t_cycles		lbpl();
	inline void		bra();
	inline t_cycles		lbra();
	inline void		brn();
	inline t_cycles		lbrn();
	inline void		bsr();
	inline t_cycles		lbsr();
	inline void		bvc();
	inline t_cycles		lbvc();
	inline void		bvs();
	inline t_cycles		lbvs();
	inline void		bit(Byte reg, Byte operand);
	inline void		clr(Byte &reg);
	inline void		clr(Word addr);
	inline void		cmp(Byte reg, Byte operand);
	inline void		cmp(Word reg, Word operand);
	inline void		com(Byte &reg);
	inline void		com(Word addr);
	inline void		daa();
	inline void		dec(Byte &reg);
	inline void		dec(Word addr);
	inline void		eor(Byte &reg, Byte operand);
	inline void		exg();
	inline void		inc(Word addr);
	inline void		inc(Byte &reg);
	inline void		jsr(Word addr);
	inline void		ld(Byte &reg, Byte operand);
	inline void		ld(Word &reg, Word operand);
	inline void		lea(Word &reg, Word addr); 
	inline void		lea_nocc(Word &reg, Word addr); 
	inline void		lsl(Byte &reg);
	inline void		lsr(Byte &reg);
	inline void		lsr(Word addr);
	inline void		lsl(Word addr);
	inline void		mul();
	inline void		jmp(Word addr);
	inline void		neg(Byte &reg);
	inline void		neg(Word addr);
	inline void		tst(Byte &reg);
	inline void		tst(Word addr);
	inline void		nop();
	inline void		or_(Byte &reg, Byte operand);
	inline void		orcc(Byte operand);
	inline t_cycles		psh(Byte what, Word &s, Word &u);
	inline t_cycles		pul(Byte what, Word &s, Word &u);
	inline void		rol(Byte &reg);
	inline void		rol(Word addr);
	inline void		ror(Byte &reg);
	inline void		ror(Word addr);
	inline t_cycles		rti();
	inline void		rts();
	inline void		sex();
	inline void		st(Byte &reg, Word addr);
	inline void		st(Word &reg, Word addr);
	inline void		adc(Byte &reg, Byte operand);
	inline void 		add(Byte &reg, Byte operand);
	inline void 		add(Word &reg, Word operand);
	inline void		sbc(Byte &reg, Byte operand);
	inline void 		sub(Byte &reg, Byte operand);
	inline void 		sub(Word &reg, Word operand);
	inline void		swi(), swi2(), swi3();
	inline void		cwai(), sync();
	inline void		tfr();

	inline void		do_br(int);
	inline t_cycles		do_lbr(int);

	// additional undocumented instructions
	void			rst();
	void			negcom(Word addr);
	void			clr1(Byte &reg);
#endif
	inline void		nmi(bool save_state);
	inline void		firq(bool save_state);
	inline void		irq(bool save_state);
	void			invalid(const char *pmessage);
	bool			use_undocumented;
public:
	void			set_use_undocumented(bool b);
	bool			is_use_undocumented() { return use_undocumented; };

// Functions to handle processor state machine
public:
	Byte			run(Word mode);
	void			do_reset(void);
	void			exit_run(void);
protected:
	Byte			runloop();

// interrupt handling:
public:
	void			reset(void);		// CPU reset
	t_cycles		exec_irqs(bool save_state = true);
	void			set_nmi(void);
	void			set_firq(void);
	void			set_irq(void);

// functions to load hexfile

private:

	void			load_intelhex(FILE *fp);
	void			load_motorola_srec(FILE *fp);
public:
	bool			load_hexfile(const char *filename,
					bool ignore_errors = false);

// cycle count
public:		
		void		set_required_cyclecount(t_cycles x_cycles);

protected:
		t_cycles	cycles;		// cycle cnt for one timer tick
		t_cycles	required_cyclecount;//cycle count for freq ctrl

// breakpoint support
protected:
		unsigned int	bp[3];
public:
		void		set_bp(int which, Word address);
		unsigned int	get_bp(int which);
		int		is_bp_set(int which);
		void		reset_bp(int which);

// interface to other classes
public:
		void		set_disassembler(Da6809 *x_da);
		void		set_status(struct s_cpu_status *stat);
		void		set_serpar(Byte b);
		uint64_t	get_cycles(bool reset = false);
		bool		set_logfile(const struct s_cpu_logfile *lf);
protected:
		int		Disassemble(Word address, DWord *pFlags,
					char **pb1, char **pb2);
		FILE		*log_fp;
		bool		do_logging;
		struct s_cpu_logfile lfs;

//****************************************************
// the memory interface is beginning from here
// it is included into the cpu class because of 
// performance 
//****************************************************

// Internal registers

private:
	Byte		*ppage[16];
	Byte		 memory[0x10000L];
	Byte		 *video_ram;
	int		 io_base_addr;
	int		 io_initialized;
	IoDevice	*ioDevices[MAX_IO_DEVICES];
	struct sIoSelect *ppIo;

public:
	// interface to video display
	Word		 isHiMem;
	Byte		*vram_ptrs[MAX_VRAM];
	Word		 changed[YBLOCKS];

private:
	void	init_memory(int himem);
	void	uninit_memory(void);

// Initialisation functions

public:

	void	initialize_io_page(Word base_addr);
	Byte	add_io_device(
			IoDevice *device,
			Word base_addr1, Byte range1,
			Word base_addr2, Byte range2);

// memory interface
public:
	void	reset_io(void);
	void	switch_mmu(Word offset, Byte val);
protected:
	void		write_rom(Word addr, Byte val);
	inline void	write(Word addr, Byte val);
	inline void	write_word(Word addr, Word val);
	inline Byte	read(const Word addr);
	inline Byte	readFromIo(Word addr);
	inline Word	read_word(Word addr);

// interface for io device protocol
public:

	virtual Byte	readIo(Word addr);
	virtual void	writeIo(Word addr, Byte val);

// Public constructor and destructor
public:
		Mc6809(int himem = 0);
	virtual	~Mc6809();
};
#endif // __mc6809_h__

