/*
    pia2.cpp


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997-2003  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <misc1.h>

#ifdef HAVE_X11
#include <X11/X.h>
#endif
#include "inout.h"
#include "pia2.h"
#include "mc6809.h"
#include "bjoystck.h"


Pia2::Pia2(Inout* x_io, Mc6809* x_cpu)
{
	cpu	 = x_cpu;
	io       = x_io;
	factor = 1.0;

#ifdef LINUX_JOYSTICK_IS_PRESENT
	joystick = new BJoystick(0);
#endif
#ifdef WIN32
	int	counter = 0;
	DWORD ticks;
	//char str[64];

	// first synchronize to begin of next tick
	while (GetTickCount() == GetTickCount()) ;
	ticks = GetTickCount();
	// loop for 100 ms
	while (GetTickCount() - ticks < 100)
		counter++;
	factor =  (1000000.0 / counter) * 3.9;
	//sprintf(str, "factor: %f, counter: %ld\n", factor, counter);
	//MessageBox(NULL, str, PROGRAMNAME " ...", MB_OK | MB_ICONWARNING);
	if (factor > 1.0)
		factor = 1.0;
#endif
}

Pia2::~Pia2()
{
#ifdef LINUX_JOYSTICK_IS_PRESENT
	delete joystick;
#endif
}

void Pia2::resetIo(void)
{
	Mc6821::resetIo();
	io->reset_joystick();
}

void Pia2::writeOutputB(Byte val)
{
	if (val & 0x40)
		io->set_bell(0);
}


Byte Pia2::readInputB(void)
{
	unsigned int buttonMask;
	int deltaX, deltaY;

	io->get_joystick(&deltaX, &deltaY, &buttonMask);

	orb &= 0xc1;
	if (buttonMask & L_MB)
		if (buttonMask & SHIFT_KEY) // shift L_MB to emulate M_MB
			orb |= 0x20;
		else
			orb |= 0x02;
	if (buttonMask & M_MB)
		if (buttonMask & SHIFT_KEY)
			orb |= 0x08;
		else
			if (buttonMask & CONTROL_KEY)
				orb |= 0x10;
			else
				orb |= 0x20;
	if (buttonMask & R_MB)
		if (buttonMask & SHIFT_KEY) // shift R_MB to emulate M_MB
			orb |= 0x20;
		else
			orb |= 0x04;
	// joystick input will either be emulated with mouse buttons
	// or with a real joystick (only linux support at the moment):
#ifdef LINUX_JOYSTICK_IS_PRESENT
	if (joystick->IsOpened()) {
		joystick->Actualize();
		if (joystick->XAxis() < -8)
			orb |= 0x02;	// joystick to left side
		if (joystick->XAxis() > 8)
			orb |= 0x04;	// joystick to left side
		if (joystick->YAxis() < -8)
			orb |= 0x10;	// joystick up
		if (joystick->YAxis() > 8)
			orb |= 0x08;	// joystick down
		if (joystick->IsButtonSet(0))
			orb |= 0x20;	// joystick "shoot"
	}
#endif
	// the emulation of the analog Eltec Joystick is implemented
	// as follows:
	// If there is no pointer movement a middle frequency of
	// about 2 KHz will be generated.
	// Pointer moves are available in io->deltaX/Y
	// if there is a move in any direction a higher/lower frequency
	// will be generated at the pia port bits 0 or 7
	// due to only three states lower/middle/higher frequency
	// an higher mouse move results in a slower joystick movement
	static Word modulX = 15;
	static Word modulY = 15;
	static Word countX = 0;
	static Word countY = 0;
	int dX, dY;
	int temp;

	countX = ++countX % modulX;
	if (!countX) {
		dX = deltaX;
		temp = 0;
		if (dX > 0)
			temp = (int)(9 * factor);
		else
			if (dX < 0)
				temp = (int)(-50 * factor);
		modulX = 15 - temp;
		countX = 0;
		orb ^= 0x01;
	}
	countY = ++countY % modulY;
	if (!countY) {
		dY = deltaY;
		temp = 0;
		if (dY > 0)
			temp = (int)(9 * factor);
		else
			if (dY < 0)
				temp = (int)(-50 * factor);
		modulY = 15 - temp;
		countY = 0;
		orb ^= 0x80;
	}
	return orb;
} // readInputB

