/*
    win32gui.cpp  user interface for Win32


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997-2003  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <misc1.h>

#ifdef WIN32

#include <ctype.h>
#include <new>
#include <math.h>
#include <crtdbg.h>
#include "resource.h"
#include "win32gui.h"
#include "inout.h"
#include "e2video.h"
#include "schedule.h"
#include "csetfreq.h"


Win32Gui *wgui = NULL;


extern LRESULT CALLBACK cpuWindowWndProc(
	HWND hwnd,
	UINT message,
	WPARAM wParam,
	LPARAM lParam);

// process window functions
// ATTENTION: all functions must be reentrant!

LRESULT CALLBACK e2windowWndProc(
	HWND hwnd,
	UINT message,
	WPARAM wParam,
	LPARAM lParam)
{
	switch(message) {
		case WM_PAINT:   wgui->onPaint(hwnd);
			// must call DefWindowProc here, don't know why?
			return DefWindowProc(hwnd, message,
				wParam, lParam);
		case WM_COMMAND: wgui->onCommand(hwnd,LOWORD(wParam));
			break;
		case WM_CHAR:	wgui->onChar(hwnd,
			(SWord)LOBYTE(wParam), (int)LOWORD(lParam));
			break;
		case WM_KEYDOWN: wgui->onKeyDown(hwnd,
				wParam, (int)LOWORD(lParam));
			break;
		case WM_TIMER:   wgui->onTimer(hwnd, (UINT)wParam);
			break;
		case WM_SIZE:    return wgui->onSize(hwnd, (int)wParam,
					(int)LOWORD(lParam), (int)HIWORD(lParam));
		case WM_DESTROY: wgui->onDestroy(hwnd);
			break;
		case WM_CLOSE: wgui->onClose(hwnd);
			break;
		case WM_ACTIVATE: wgui->onActivate(hwnd, (WORD)LOWORD(wParam), (HWND)lParam);
			break;
		case WM_GETMINMAXINFO: return wgui->onMinMaxInfo((MINMAXINFO *)lParam);
		case WM_PALETTECHANGED:
			wgui->onPaletteChanged(hwnd);
			break;
		case WM_MOUSEMOVE:
			wgui->onMouseMove(hwnd, LOWORD(lParam), HIWORD(lParam), wParam);
			break;
		default: return DefWindowProc(hwnd, message, wParam, lParam);
	} // switch

	return TRUE;
} // e2windowWndProc 

void Win32Gui::resetMouseMoveCoords()
{
	io->put_joystick(0, 0);
}

void Win32Gui::onTimer(HWND hwnd, UINT id)
{
	if (id == idTimer)
	{
		unsigned int newButtonMask;
		short display_block;
		HDC hdc;

		// check if CPU view has to be updated
		if (schedy->get_status(cpu_stat))
		{
			bool is_running = (cpu_stat.state == S_RUN ||
					cpu_stat.state == S_NEXT);
			CheckMenuItem(menu2, IDM_RUN,
				MF_BYCOMMAND | (is_running ? MF_CHECKED : MF_UNCHECKED)); 
			CheckMenuItem(menu2, IDM_STOP,
				MF_BYCOMMAND | (!is_running ? MF_CHECKED : MF_UNCHECKED)); 
			update_cpuview(cpu_stat);
			if (cpu_stat.state == S_INVALID)
			{
				char err_msg[128];

        		sprintf((char *)&err_msg, "\
Got invalid instrucion\n\
pc=%04x instr=%02x %02x %02x %02x\n\
Processor stopped. To\n\
continue press Reset button",
					cpu_stat.pc,
					cpu_stat.instruction[0], cpu_stat.instruction[1],
					cpu_stat.instruction[2], cpu_stat.instruction[3]);
        			popup_message(err_msg);
			}
		}

		// update graphic display (only if display memory has changed)
		hdc = GetDC(hwnd);
		for (display_block = 0; display_block < YBLOCKS; display_block++)
			if (cpu->changed[display_block])
				update_block(display_block, hdc);
		ReleaseDC(hwnd, hdc);	

		if (GetActiveWindow() == e2screen) {
			newButtonMask = 0;
			if (IsLButtonDown())
				newButtonMask |= L_MB;
			if (IsMButtonDown())
				newButtonMask |= M_MB;
			if (IsRButtonDown())
				newButtonMask |= R_MB;
			if (GetKeyState(VK_CONTROL) < 0)
				newButtonMask |= CONTROL_KEY;
			if (GetKeyState(VK_SHIFT) < 0)
				newButtonMask |= SHIFT_KEY;
			io->put_joystick(newButtonMask);
		}
		// check if program can be savely shut down
		if (schedy->is_finished())
		{
			CloseApp(hwnd);
		}
	}
} // onTimer

void Win32Gui::onCommand(HWND hwnd, int cmd)
{
	switch (cmd) {
		// Main Window menu
		case IDM_EXIT:          CloseApp(hwnd);				break;
		case IDM_RUN:           set_new_state(S_RUN);       break;
		case IDM_STOP:          set_new_state(S_STOP);      break;
		case IDM_RESET:         set_new_state(S_RESET_RUN); break;
		case IDM_VIEW:          toggle_cpu();               break;
		case IDM_BREAKPOINTS:   popup_bp(hwnd);             break;
		case IDM_LOGFILE:       popup_log(hwnd);            break;
		case IDM_FREQUENCY0:    toggle_freqency();          break;
		case IDM_UNDOCUMENTED:  toggle_undocumented();      break;
		case IDM_ABOUT:         popup_about(hwnd);          break;
		case IDM_COPYRIGHT:     popup_copyright(hwnd);      break;
		case IDM_INTRODUCTION:	popup_help(hwnd);           break;
	}
	return;
} // onCommand

void Win32Gui::onChar(HWND hwnd, SWord ch, int repeat)
{
	SWord key;
	
	if ((key = translate_to_ascii(ch)) >= 0)
	io->put_ch((Byte)key);

} // onChar

void Win32Gui::onKeyDown(HWND hwnd, SWord ch, int repeat)
{
	SWord key;
	
	if ((key = translate_to_ascii1(ch)) >= 0)
	io->put_ch((Byte)key);

} // onKeyDown

void Win32Gui::onPaletteChanged(HWND hwnd)
{
	HDC hdc;

	hdc = GetDC(e2screen);
	UpdateColors(hdc);
	ReleaseDC(e2screen, hdc);
} // onPaletteChanged

void Win32Gui::onMouseMove(HWND hwnd, Word newX, Word newY, Word fwKeys)
{
	static int oldX;
	static int oldY;
	static Word init = 1;

	if (init) {
		oldX = newX;
		oldY = newY;
		init = 0;
		return;
	}
	io->put_joystick(newX - oldX, newY - oldY);
	oldX = newX;
	oldY = newY;
	mouseMoved = 1;
}

void Win32Gui::onPaint(HWND hwnd)
{
	int			display_block;
	HDC			hdc;
	PAINTSTRUCT	ps;

	// On Windows Request to update a part or the whole Client rect
	// Always update everything
	hdc = BeginPaint(hwnd, &ps);
	e2video->init_blocks_to_update();
	for (display_block = 0; display_block < YBLOCKS; display_block++)
		update_block(display_block, hdc);
	EndPaint(hwnd, &ps);

} // onPaint

int Win32Gui::onSize(HWND hwnd, int sizeType, int width, int height)
{
	RECT    oldRect, rect;

	if (sizeType == SIZE_RESTORED) {
		if (width % WINDOWWIDTH != 0 || height % WINDOWHEIGHT != 0) {
			guiXSize = (width + (WINDOWWIDTH / 2)) /  WINDOWWIDTH;
			guiYSize = (height + (WINDOWHEIGHT / 2)) /  WINDOWHEIGHT;
			GetWindowRect(e2screen, &oldRect);
			rect.left       = (long)0;
			rect.top        = (long)0;
			rect.right      = (long)WINDOWWIDTH * guiXSize;
			rect.bottom     = (long)WINDOWHEIGHT * guiYSize;
			if (AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, TRUE, 0)) {
				width  = rect.right - rect.left;
				height = rect.bottom - rect.top;
			};
			MoveWindow(e2screen, oldRect.left, oldRect.top,
				width, height, TRUE);
		}
	}
	return  0;
} // onSize

// Savely close the Application
void Win32Gui::CloseApp(HWND hwnd)
{
	set_new_state(S_EXIT);
	KillTimer(hwnd, idTimer);
	idTimer = 0;
	if (menubar != NULL)
		DestroyMenu(menubar);
	if (hwnd != NULL)
		DestroyWindow(hwnd);
} // CloseApp


void Win32Gui::onClose(HWND hwnd)
{
	CloseApp(hwnd); 
} // onClose

void Win32Gui::onActivate(HWND hwnd, WORD what, HWND hwnd1)
{
	if (what == WA_ACTIVE || what == WA_CLICKACTIVE)
		e2video->init_blocks_to_update();
}

void Win32Gui::onDestroy(HWND)
{
	PostQuitMessage(0);
} // onDestroy

int Win32Gui::onMinMaxInfo(MINMAXINFO *lpmmi)
{
	RECT    rect;

	rect.left       = (long)0;
	rect.top        = (long)0;
	rect.right      = (long)WINDOWWIDTH;
	rect.bottom     = (long)WINDOWHEIGHT;
	if (AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, TRUE, 0)) {
		lpmmi->ptMinTrackSize.x = rect.right - rect.left;
		lpmmi->ptMinTrackSize.y = rect.bottom - rect.top;
	};
	rect.right      = (long)WINDOWWIDTH * MAX_GUIXSIZE;
	rect.bottom     = (long)WINDOWHEIGHT * MAX_GUIYSIZE;
	if (AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, TRUE, 0)) {
		lpmmi->ptMaxTrackSize.x = rect.right - rect.left;
		lpmmi->ptMaxTrackSize.y = rect.bottom - rect.top;
	};
	return  0;
} // onMinMaxInfo

SWord Win32Gui::translate_to_ascii(SWord key)
{
	switch (key) {
		case 0xdf: return '^';  // german sharp s
		case 0xa7: return '#';  // paragraph
		case 0xb4: return '\'';
		case 0xb5: return 'u';  // my
		case 0xb0: return '0';  // degree
		case 0xb2: return '2';  // index 2
		case 0xb3: return '3';  // index 3
		case 0xf6: return '[';  // german umlaut oe
		case 0xe4: return ']';  // german umlaut ae
		case 0xfc: return '\\'; // german umlaut ue
		case 0xd6: return '{';  // german umlaut OE
		case 0xc4: return '}';  // german umlaut AE
		case 0xdc: return '|';  // german umlaut UE
	}
	return key;
} // translate_to_ascii

SWord Win32Gui::translate_to_ascii1(SWord key)
{               
	if (GetKeyState(VK_CONTROL) < 0)
		switch(key) {
			case VK_PAUSE:
			case VK_CANCEL: if (cpu) cpu->set_nmi(); return -1;
			case VK_HOME:   return 0xb7;
			case VK_UP:     return 0xb8;
			case VK_PRIOR:  return 0xb9;
			case VK_LEFT:   return 0xb4;
			case VK_CLEAR:  return 0xb5;
			case VK_RIGHT:  return 0xb6;
			case VK_END:    return 0xb1;
			case VK_DOWN:   return 0xb2;
			case VK_NEXT:   return 0xb3;
			case VK_F11:	return 0xfa; // PAT09: RIGHT MOST
			case VK_F12:	return 0x91; // PAT09: LEFT  MOST
			default: return -1;
		}
	if (GetKeyState(VK_SHIFT) < 0)
		switch(key) {
			case VK_F1:     return 0xca; // PAT09: F11
			case VK_F2:     return 0xcb;
			case VK_F3:     return 0xcc;
			case VK_F4:     return 0xcd;
			case VK_F5:     return 0xce;
			case VK_F6:     return 0xcf;
			case VK_F7:     return 0xd0;
			case VK_F8:     return 0xd1;
			case VK_F9:     return 0xd2; // PAT09: F19
			case VK_F11:    return 0xea; // PAT09: RIGHT MOST
			case VK_F12:    return 0x81; // PAT09: LEFT  MOST

			case VK_PAUSE:
			case VK_CANCEL: if (cpu) cpu->set_nmi(); return -1;
			case VK_HOME:   return 0xe7;
			case VK_UP:     return 0xe8;
			case VK_PRIOR:  return 0xe9;
			case VK_LEFT:   return 0xe4;
			case VK_CLEAR:  return 0xe5;
			case VK_RIGHT:  return 0xe6;
			case VK_END:    return 0xe1;
			case VK_DOWN:   return 0xe2;
			case VK_NEXT:   return 0xe3;
			default: return -1;
		}
	switch(key) {
		case VK_F1:     return 0xc0;
		case VK_F2:     return 0xc1;
		case VK_F3:     return 0xc2;
		case VK_F4:     return 0xc3;
		case VK_F5:     return 0xc4;
		case VK_F6:     return 0xc5;
		case VK_F7:     return 0xc6;
		case VK_F8:     return 0xc7;
		case VK_F9:     return 0xc8;
		case VK_F10:    return 0xc9;
		case VK_F11:    return 0xfa; // PAT09: RIGHT MOST
		case VK_F12:    return 0x91; // PAT09: LEFT  MOST

		case VK_HOME:   return 0xf7;
		case VK_UP:     return 0xf8;
		case VK_PRIOR:  return 0xf9;
		case VK_LEFT:   return 0xf4;
		case VK_CLEAR:  return 0xf5;
		case VK_RIGHT:  return 0xf6;
		case VK_END:    return 0xf1;
		case VK_DOWN:   return 0xf2;
		case VK_NEXT:   return 0xf3;
		default: return -1;
	}
	return -1;
}

void Win32Gui::initialize(struct sGuiOptions *pOptions)
{
	AbstractGui::initialize(pOptions);
	wgui        = this;
	hAccel      = NULL;
	hAccelHwnd  = NULL;
	palette		= NULL; // needed for cpu_view
	e2screen	= NULL;
	use_colors	= !stricmp(pOptions->color, "default");
	nColors		= pOptions->nColors;
	bp_input[0] = 0;
	bp_input[1] = 0;
	lfs.logFileName[0] = '\0';
	lfs.minAddr   = 0x0000;
	lfs.maxAddr   = 0xFFFF;
	lfs.startAddr = 0x10000;
	lfs.stopAddr  = 0x10000;
	copy_block  = NULL;
	frequency_control_on = false;
	initialize_conv_tables();
	initialize_e2window(pOptions);
} // initialize

// return 0 on success
int Win32Gui::popup_help(HWND hwnd)
{
	char helpfile[PATH_MAX];
	char curdir[PATH_MAX];
	int res;

	GetCurrentDirectory(PATH_MAX, curdir);
	if (strlen(pOptions->doc_dir) != 0)
		strcpy(helpfile, pOptions->doc_dir);
	else {
		strcpy(helpfile, curdir);
		strcat(helpfile, "\\doc\\");
	}
	if (helpfile[strlen(helpfile)-1] != PATHSEPARATOR)
		strcat(helpfile, PATHSEPARATORSTRING);
	strcat(helpfile, "flexemu.htm");
	if ((res = (int)ShellExecute(
		hwnd,           // handle of parent window
		"open",         // open file
		helpfile,       // file to open
		NULL,           // parameters
		curdir,         // directory
		SW_SHOWNORMAL   // openmode for new window
	)) > 32)
		return 0; // success
	return -1; // no success
}

void Win32Gui::popup_message(char *pmessage)
{
	if (*pmessage)
		MessageBox(e2screen, pmessage, PROGRAMNAME " error",
			MB_OK | MB_ICONERROR);
}

void Win32Gui::main_loop(void)
{
	MSG msg;

	while (GetMessage(&msg, NULL, 0, 0) != 0)
	{
		if (hAccel == NULL || hAccelHwnd != msg.hwnd ||
		     !TranslateAccelerator(msg.hwnd, hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	} // while
}

void Win32Gui::event_queue(void)
{
	MSG msg;
	int count;
	int newButtonMask;
	static DWORD oldTime = 0;

	count = 0;
	mouseMoved = 0;
	// process all queued events
	while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
		if (hAccel == NULL || hAccelHwnd != msg.hwnd ||
		     !TranslateAccelerator(msg.hwnd, hAccel, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			count++;
		}
	} // while

	if (GetActiveWindow() == e2screen) {
		newButtonMask = 0;
		if (IsLButtonDown())
			newButtonMask |= L_MB;
		if (IsMButtonDown())
			newButtonMask |= M_MB;
		if (IsRButtonDown())
			newButtonMask |= R_MB;
		if (GetKeyState(VK_CONTROL) < 0)
			newButtonMask |= CONTROL_KEY;
		if (GetKeyState(VK_SHIFT) < 0)
			newButtonMask |= SHIFT_KEY;
		io->put_joystick(newButtonMask);
	}

	// hack for reset mouse move deltaX and Y to zero if
	// no mouse moving in progress
	if (mouseMoved) {
		oldTime = GetTickCount();
	} else {
		// after 40 ms reset deltaX/Y
		if ((GetTickCount() - oldTime) >= 40)
			wgui->resetMouseMoveCoords();
	}
}


void Win32Gui::set_bell(int percent)
{
	// volume in percent is ignored
	Beep(400, 200);
} // set_bell

void Win32Gui::update_block(int block_number, HDC hdc)
{
	if (e2video->are_blocks_to_update() || cpu->changed[block_number]) {
		cpu->changed[block_number] = 0;
		if (nColors > 2)
			update_color_block(block_number, hdc);
		else
			update_bw_block(block_number, hdc);
	}
}

void Win32Gui::update_bw_block(int block_number, HDC hdc)
{
	HDC	  hMemoryDC;
	HBITMAP hBitmapOrig;
	int     firstpartHeight; // Height of first part of divided block
	int     startLine;       // start scanline of block
	int     i;
	Byte    *src, *block;
	HBITMAP img;

	hMemoryDC = CreateCompatibleDC(NULL);
	
	if (!(e2video->vico1 & 0x02)) {
		// copy block from video ram into device independant bitmap
		if (e2video->vico1 & 0x01)
			src = cpu->vram_ptrs[0x08] + block_number * YBLOCK_SIZE;
		else
			src = cpu->vram_ptrs[0x0C] + block_number * YBLOCK_SIZE;
		block = copy_block;
		img = image1[guiXSize-1][guiYSize-1];
		switch ((guiXSize << 4) | guiYSize) {
		int     j;
			case 0x11: {// single width, single height
				// if standard gui Size directly copy from video_ram:
				Byte *btrg = copy_block;
				memcpy(btrg, src, RASTERLINE_SIZE * BLOCKHEIGHT);
				break; }
				// if not standard gui Size first calculate contents of
				// block to display:
			case 0x12: // single width, double height
			case 0x13: // single width, tripple height
			case 0x14: {// single width, quadrupple height
				Byte *btrg = copy_block;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < guiYSize; j++) {
						memcpy(btrg, src, RASTERLINE_SIZE);
						btrg += RASTERLINE_SIZE;
					}
					src += RASTERLINE_SIZE;
				}
				break; }
			case 0x21: {// double width, single heigth
				Word    *wtrg;
				wtrg = (Word *)copy_block;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						*(wtrg++) = conv_2byte_tab[*(src++)];
					}
				}
				break; }
			case 0x22: {// double width, double height
				Word    *wtrg;
				wtrg = (Word *)copy_block;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						*wtrg = *(wtrg+RASTERLINE_SIZE) =
						conv_2byte_tab[*(src++)];
						wtrg++;
					}
					wtrg += RASTERLINE_SIZE;
				}
				break; }
			case 0x23: { // double width, tripple height
				Word    *wtrg;
				wtrg = (Word *)copy_block;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						*wtrg = *(wtrg+RASTERLINE_SIZE) =
						  *(wtrg+(2*RASTERLINE_SIZE)) =
							conv_2byte_tab[*(src++)];
						wtrg++;
					}
					wtrg += 2*RASTERLINE_SIZE;
				}
				break; }
			case 0x24: { // double width, quadrupple height
				Word    *wtrg;
				wtrg = (Word *)copy_block;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						*wtrg = *(wtrg+RASTERLINE_SIZE) =
						  *(wtrg+(2*RASTERLINE_SIZE)) =
						  *(wtrg+(3*RASTERLINE_SIZE)) =
							conv_2byte_tab[*(src++)];
						wtrg++;
					}
					wtrg += 3*RASTERLINE_SIZE;
				}
				break; }
			case 0x31: { // triple width, single height
				Byte *btrg = copy_block;
				DWord pattern;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						pattern = conv_3byte_tab[*(src++)];
						*(btrg++) = (pattern & 0xFF0000) >> 16;
						*(btrg++) = (pattern & 0xFF00) >> 8;
						*(btrg++) = (pattern & 0xFF);
					}
				}
				break; }
			case 0x32: { // triple width, double height
				Byte *btrg = copy_block;
				DWord pattern;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						pattern = conv_3byte_tab[*(src++)];
						*btrg = *(btrg + 3 * RASTERLINE_SIZE)
							= (pattern & 0xFF0000) >> 16;
						*(btrg+1) = *(btrg + 3*RASTERLINE_SIZE + 1)
							= (pattern & 0xFF00) >> 8;
						*(btrg+2) = *(btrg + 3*RASTERLINE_SIZE + 2)
							= (pattern & 0xFF);
						btrg += 3;
					}
					btrg += 3 * RASTERLINE_SIZE;
				}
				break; }
			case 0x33: { // triple width, triple height
				Byte *btrg = copy_block;
				DWord pattern;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						pattern = conv_3byte_tab[*(src++)];
						*btrg =
						*(btrg + 3 * RASTERLINE_SIZE) =
						*(btrg + 6 * RASTERLINE_SIZE) =
							(pattern & 0xFF0000) >> 16;
						*(btrg+1) =
						*(btrg + 3*RASTERLINE_SIZE + 1) =
						*(btrg + 6*RASTERLINE_SIZE + 1) =
							(pattern & 0xFF00) >> 8;
						*(btrg+2) =
						*(btrg + 3*RASTERLINE_SIZE + 2) =
						*(btrg + 6*RASTERLINE_SIZE + 2) =
							(pattern & 0xFF);
						btrg += 3;
					}
					btrg += 6 * RASTERLINE_SIZE;
				}
				break; }
			case 0x34: { // triple width, quadruple height
				Byte *btrg = copy_block;
				DWord pattern;
				for (i = 0; i < BLOCKHEIGHT; i++) {
					for (j = 0; j < RASTERLINE_SIZE; j++) {
						pattern = conv_3byte_tab[*(src++)];
						*btrg =
						*(btrg + 3 * RASTERLINE_SIZE) =
						*(btrg + 6 * RASTERLINE_SIZE) =
						*(btrg + 9 * RASTERLINE_SIZE) =
							(pattern & 0xFF0000) >> 16;
						*(btrg+1) =
						*(btrg + 3*RASTERLINE_SIZE + 1) =
						*(btrg + 6*RASTERLINE_SIZE + 1) =
						*(btrg + 9*RASTERLINE_SIZE + 1) =
							(pattern & 0xFF00) >> 8;
						*(btrg+2) =
						*(btrg + 3*RASTERLINE_SIZE + 2) =
						*(btrg + 6*RASTERLINE_SIZE + 2) =
						*(btrg + 9*RASTERLINE_SIZE + 2) =
							(pattern & 0xFF);
						btrg += 3;
					}
					btrg += 9 * RASTERLINE_SIZE;
				}
				break; }
					default:
						img = image1[0][0];
				} // switch
		SetDIBits(
			hdc, img,
			0, BLOCKHEIGHT * guiYSize, 
			block, bmi1[guiXSize-1][guiYSize-1], DIB_RGB_COLORS);

		hBitmapOrig = SelectBitmap(hMemoryDC, img);
		startLine = ((WINDOWHEIGHT - e2video->vico2 +
			block_number * BLOCKHEIGHT) % WINDOWHEIGHT) * guiYSize;
		if (block_number == e2video->divided_block) {
			firstpartHeight = e2video->vico2 % BLOCKHEIGHT;
			// first half display on the bottom of the window
			BitBlt(hdc, 0, startLine,
				BLOCKWIDTH * guiXSize,
				firstpartHeight * guiYSize,
				hMemoryDC, 0, 0,
				SRCCOPY);
			// second half display on the top of window
			BitBlt(hdc, 0, 0,
				BLOCKWIDTH * guiXSize,
				(BLOCKHEIGHT - firstpartHeight) * guiYSize,
				hMemoryDC, 0, firstpartHeight * guiYSize,
				SRCCOPY);
			// first half display on the bottom of the window
		} else {
			BitBlt(hdc, 0, startLine,
				BLOCKWIDTH * guiXSize, BLOCKHEIGHT * guiYSize,
				hMemoryDC, 0, 0, SRCCOPY);
		} // else
		SelectBitmap(hMemoryDC, hBitmapOrig);
	} else {
		// display an "empty" screen:
		for (i = 0; i < RASTERLINE_SIZE * guiXSize * BLOCKHEIGHT * guiYSize; i++) {
			copy_block[i] = 0xff;
		}
		SetDIBits(
			hdc, image1[guiXSize-1][guiYSize-1],
			0, BLOCKHEIGHT * guiYSize, 
			copy_block, bmi1[guiXSize-1][guiYSize-1], DIB_RGB_COLORS);
		hBitmapOrig = SelectBitmap(hMemoryDC, image1[guiXSize-1][guiYSize-1]);
		BitBlt(hdc, 0,
			(block_number * BLOCKHEIGHT) * guiYSize,
			BLOCKWIDTH * guiXSize,
			BLOCKHEIGHT * guiYSize,
			hMemoryDC, 0, 0, SRCCOPY);
		SelectBitmap(hMemoryDC, hBitmapOrig);
	} // else
	DeleteDC(hMemoryDC);
} // update_bw_block

void Win32Gui::update_color_block(int block_number, HDC hdc)
{
	HDC	  hMemoryDC;
	HBITMAP hBitmapOrig;
	int     firstpartHeight; // Height of first part of divided block
	int     startLine;       // start scanline of block
	int     i;
	Byte    *src, *block;
	HBITMAP img;
	HPALETTE oldPalette;

	hdc = GetDC(e2screen);
	hMemoryDC = CreateCompatibleDC(NULL);
	oldPalette = SelectPalette(hdc, palette, TRUE);
	RealizePalette(hdc);
	if (!(e2video->vico1 & 0x02)) {
		// copy block from video ram into device independant bitmap
		if (e2video->vico1 & 0x01)
			src = cpu->vram_ptrs[0x08] + block_number * YBLOCK_SIZE;
		else
			src = cpu->vram_ptrs[0x0C] + block_number * YBLOCK_SIZE;
		block = copy_block;
		img = image6[guiXSize-1][guiYSize-1];
		CopyToZPixmap(block_number, copy_block, src, 8, (unsigned long *)pen);

		SetDIBits(
			hdc, img,
			0, BLOCKHEIGHT * guiYSize, 
			block, bmi6[guiXSize-1][guiYSize-1], DIB_PAL_COLORS);
		hBitmapOrig = SelectBitmap(hMemoryDC, img);
		startLine = ((WINDOWHEIGHT - e2video->vico2 +
			block_number * BLOCKHEIGHT) % WINDOWHEIGHT) * guiYSize;
		if (block_number == e2video->divided_block) {
			firstpartHeight = e2video->vico2 % BLOCKHEIGHT;
			// first half display on the bottom of the window
			BitBlt(hdc, 0, startLine,
				BLOCKWIDTH * guiXSize,
				firstpartHeight * guiYSize,
				hMemoryDC, 0, 0,
				SRCCOPY);
			// second half display on the top of window
			BitBlt(hdc, 0, 0,
				BLOCKWIDTH * guiXSize,
				(BLOCKHEIGHT - firstpartHeight) * guiYSize,
				hMemoryDC, 0, firstpartHeight * guiYSize,
				SRCCOPY);
			// first half display on the bottom of the window
		} else {
			BitBlt(hdc, 0, startLine,
				BLOCKWIDTH * guiXSize, BLOCKHEIGHT * guiYSize,
				hMemoryDC, 0, 0, SRCCOPY);
		} // else
		SelectBitmap(hMemoryDC, hBitmapOrig);
	} else {
		// display an "empty" screen:
		for (i = 0; i < RASTERLINE_SIZE * guiXSize * BLOCKHEIGHT * guiYSize; i++) {
			copy_block[i] = 0xff;
		}
		SetDIBits(
			hdc, image6[guiXSize-1][guiYSize-1],
			0, BLOCKHEIGHT * guiYSize, 
			copy_block, bmi6[guiXSize-1][guiYSize-1], DIB_PAL_COLORS);
		hBitmapOrig = SelectBitmap(hMemoryDC, image6[guiXSize-1][guiYSize-1]);
		BitBlt(hdc, 0,
			(block_number * BLOCKHEIGHT) * guiYSize,
			BLOCKWIDTH * guiXSize,
			BLOCKHEIGHT * guiYSize,
			hMemoryDC, 0, 0, SRCCOPY);
		SelectBitmap(hMemoryDC, hBitmapOrig);
	} // else
	SelectPalette(hdc, oldPalette, FALSE);
	DeleteDC(hMemoryDC);
} // update_color_block

void Win32Gui::initialize_e2window(struct sGuiOptions *pOptions)
{
	HWND w;

	if (!registerWindowClasses(pOptions->hInstance, 0)) {
		MessageBox(NULL, "RegisterClassEx failed\n" \
			"Unable to create Mainwindow of " PROGRAMNAME,
			PROGRAMNAME " error",
			MB_OK | MB_ICONERROR);
		exit(EXIT_FAILURE);
	}
	if ((w = create_main_view(pOptions)) == NULL) {
		MessageBox(NULL, "CreateWindowEx failed\n" \
			"Unable to create Mainwindow of " PROGRAMNAME,
			PROGRAMNAME " error",
			MB_OK | MB_ICONERROR);
		exit(EXIT_FAILURE);
	}       
	create_cpuview(w, pOptions);
	initialize_after_create(w, pOptions);
	manage_widget(w, pOptions);
	initialize_after_open(w, pOptions);
	e2screen = w;
} // initialize_e2window


BOOL Win32Gui::registerWindowClasses (HINSTANCE hinst, UINT ResPoolID)
{
	WNDCLASSEX      wcex ;

	// Fill in window class structure with parameters that describe
	// the Main window.
	wcex.cbSize        = sizeof (WNDCLASSEX) ;
	wcex.style         = CS_BYTEALIGNCLIENT | CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS ;
	wcex.lpfnWndProc   = e2windowWndProc ;
	wcex.cbClsExtra    = 0 ;
	wcex.cbWndExtra    = 0 ;
	wcex.hInstance     = hinst ;
	wcex.hIcon         = LoadIcon (hinst, MAKEINTRESOURCE (IDI_AFLEXMAIN)) ;
	wcex.hCursor       = LoadCursor (NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
//      wcex.lpszMenuName  = MAKEINTRESOURCE(IDR_MAINMENU);
	wcex.lpszMenuName  = NULL;
	wcex.lpszClassName = "Flexemu";
	wcex.hIconSm       = NULL ;
	if (!RegisterClassEx (&wcex))
		return FALSE;

	// CPU Window
	wcex.lpfnWndProc   = cpuWindowWndProc ;
	wcex.hIcon         = LoadIcon (hinst, MAKEINTRESOURCE (IDI_FLEXCPU)) ;
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszClassName = "Mc6809";
	if (!RegisterClassEx (&wcex))
		return FALSE;
	return TRUE;
}

HWND Win32Gui::create_main_view(struct sGuiOptions *pOptions)
{
   HWND hwnd;
   RECT rect;
   int  width, height;
   int  count;

   MENUITEMINFO menuItem = {
	sizeof(MENUITEMINFO),           // size of structure
	MIIM_ID | MIIM_STATE | MIIM_TYPE, // set ID, state and type
	MFT_STRING,                     // menutype (String, Bitmap, Sep. ...)
	MFS_ENABLED,                    // menustate (enabled, checked ...)
	0,                              // ID of menuitem
	NULL,                           // handle of submenu
	NULL,                           // bitmap for checked
	NULL,                           // bitmap for unchecked
	0,                              // user defined data
	"",                             // item string ...
	0};                             // stringlength if returned

   hInstance = pOptions->hInstance;
   menubar = CreateMenu();

   if (menubar != NULL) {
	menu1 = CreateMenu();
	AppendMenu(menubar, MF_POPUP | MF_STRING, (UINT)menu1, "&File");
	menu2 = CreateMenu();
	AppendMenu(menubar, MF_POPUP | MF_STRING, (UINT)menu2, "&Processor");
	menu3 = CreateMenu();
	AppendMenu(menubar, MF_POPUP | MF_STRING, (UINT)menu3, "&Help");

	menuItem.dwTypeData = "&Exit";
	menuItem.wID = IDM_EXIT;
	InsertMenuItem(menu1, 1, TRUE, &menuItem);

	count = 1;
	menuItem.dwTypeData = "&Run";
	menuItem.wID = IDM_RUN;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&Stop";
	menuItem.wID = IDM_STOP;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "R&eset";
	menuItem.wID = IDM_RESET;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.fType = MFT_SEPARATOR;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&View...";
	menuItem.fType = MFT_STRING;
	menuItem.wID = IDM_VIEW;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&Breakpoints...";
	menuItem.fType = MFT_STRING;
	menuItem.wID = IDM_BREAKPOINTS;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&Logging...";
	menuItem.fType = MFT_STRING;
	menuItem.wID = IDM_LOGFILE;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.fType = MFT_SEPARATOR;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&Frequency 1.3396 MHz";
	menuItem.fType = MFT_STRING;
	menuItem.wID = IDM_FREQUENCY0;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&Undocumented Instructions";
	menuItem.fType = MFT_STRING;
	menuItem.wID = IDM_UNDOCUMENTED;
	InsertMenuItem(menu2, count++, TRUE, &menuItem);

	count = 1;
	menuItem.dwTypeData = "&Introduction";
	menuItem.wID = IDM_INTRODUCTION;
	InsertMenuItem(menu3, count++, TRUE, &menuItem);      
	menuItem.dwTypeData = "&About " PROGRAMNAME;
	menuItem.wID = IDM_ABOUT;
	InsertMenuItem(menu3, count++, TRUE, &menuItem);
	menuItem.dwTypeData = "&Licence";
	menuItem.wID = IDM_COPYRIGHT;
	InsertMenuItem(menu3, count++, TRUE, &menuItem);
   }

   rect.left    = (long)0;
   rect.right   = (long)WINDOWWIDTH * guiXSize;
   rect.top     = (long)0;
   rect.bottom  = (long)WINDOWHEIGHT * guiYSize;
   if (AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, TRUE, 0)) {
	width  = rect.right - rect.left;
	height = rect.bottom - rect.top;
   } else {
	width  = WINDOWWIDTH+8;
	height = WINDOWHEIGHT+24;
   }
   hwnd = CreateWindowEx (0,            // Extended window styles
		    PROGRAMNAME,        // Address of registered class name
		    PROGRAMNAME,          // Address of window name
		    WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU |
		       WS_THICKFRAME | WS_MINIMIZEBOX,  // Window style
		    CW_USEDEFAULT,      // Horizontal position of window
		    0,                  // Vertical position of window
		    width,              // Window width
		    height,             // Window height
		    NULL,               // Handle of parent or owner window
		    menubar,            // Handle of menu for this window
		    pOptions->hInstance,// Handle of application instance
		    NULL) ;             // Address of window-creation data

    _ASSERT (NULL != hwnd) ;

    if (!hwnd)
       return NULL ;
    return hwnd ;
}  // create_main_view


BOOL CALLBACK aboutDialogProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message) {
		case WM_INITDIALOG:
			return wgui->onAboutInit(hwnd);
		case WM_COMMAND:
			return wgui->onAboutCommand(hwnd,LOWORD(wParam));
		case WM_CLOSE:
			return wgui->onAboutClose(hwnd);
	} // switch
	return FALSE;
} // aboutDialogProc

void Win32Gui::toggle_undocumented(void)
{
	is_use_undocumented = !is_use_undocumented;
	cpu->set_use_undocumented(is_use_undocumented);
	CheckMenuItem(menu2, IDM_UNDOCUMENTED,
		MF_BYCOMMAND | (is_use_undocumented ? MF_CHECKED : MF_UNCHECKED)); 
}

void Win32Gui::toggle_freqency()
{
	float frequency;

	frequency_control_on = !frequency_control_on;
	frequency = frequency_control_on ? 1.3396f : 0.0f;
	schedy->sync_exec(new CSetFrequency(*schedy, frequency));
	CheckMenuItem(menu2, IDM_FREQUENCY0,
		MF_BYCOMMAND | (frequency_control_on ? MF_CHECKED : MF_UNCHECKED)); 
}

void Win32Gui::popup_about(HWND hwnd)
{
	DialogBox(hInstance, MAKEINTRESOURCE(IDD_AB_DIALOG), hwnd, CALLBACKCAST aboutDialogProc);
} // popup_about

void Win32Gui::popdown_about(HWND hwnd)
{
	EndDialog(hwnd, 0);
}

void Win32Gui::stripBlanks(char *str)
{
	char *p1, *p2;

	p1 = str;
	p2 = str;
	while (*p2) {
		// skip all leading blanks
		while (*p2 == ' ')
			p2++;
		// copy line
		while (*p2 && *p2 != '\n')
			*(p1++) = *(p2++);
		*(p1++) = *p2;
		if (p2)
			p2++;
	} // while
	*p1 = '\0';
} // stripBlanks

BOOL Win32Gui::onAboutInit(HWND hwnd)
{
	char    aboutstring[512];


	memset(aboutstring, 0, 512);
	strcpy(aboutstring, "About ");
	strcat(aboutstring, PROGRAMNAME);
	SetWindowText(hwnd, aboutstring);
	strcpy(aboutstring, HEADER1);
	strcat(aboutstring, "V " PROGRAM_VERSION);
	strcat(aboutstring, HEADER2);
	stripBlanks(aboutstring);
	SetDlgItemText(hwnd, IDC_AB_TEXT, aboutstring);
	return TRUE;
}

BOOL Win32Gui::onAboutClose(HWND hwnd)
{
	popdown_about(hwnd);
	return TRUE;
}

BOOL Win32Gui::onAboutCommand(HWND hwnd, int cmd)
{
	switch (cmd) {
		// BP-Dialog button controls
		case IDC_AB_OK:         popdown_about(hwnd); break;
		case IDC_AB_COPYRIGHT:  popup_copyright(hwnd);
					popdown_about(hwnd); break;
		default:                return FALSE;
	}
	return TRUE;
} // onAboutCommand

void Win32Gui::popup_copyright(HWND hwnd)
{
	char str[PATH_MAX];
	char *copyrightFile = "COPYING.TXT";

	GetCurrentDirectory(PATH_MAX, str);
	if ((int)ShellExecute(
		hwnd,           // handle of parent window
		"open",         // open file
		copyrightFile,  // file to open
		NULL,           // parameters
		str,            // directory
		SW_SHOWNORMAL   // openmode for new window
		) <= 32) {
		sprintf(str, "Unable to display file %s", copyrightFile);
		MessageBox(e2screen, str, PROGRAMNAME " error",
			MB_OK | MB_ICONERROR);
	}
} // popup_copyright

int Win32Gui::CheckDeviceSupport(HDC aHdc, Word modifyOptions, Word *nrOfColors)
{
	int cNumColors, rasterCaps;

	rasterCaps = GetDeviceCaps(aHdc, RASTERCAPS);
	if (!(rasterCaps & RC_BITBLT) || !(rasterCaps & RC_DI_BITMAP)) {
		if (!modifyOptions)
			return 0; // fatal error, no BitBlt or SetDIBits available
		popup_message("Device capabilities insufficient to start this application\n\
(no BitBlt or SetDIBits available)");
		exit(1);
	}
	cNumColors = GetDeviceCaps(aHdc, NUMCOLORS);
	if (cNumColors < 0)
		return 1;  // success: more than 256 colors
	if (cNumColors < *nrOfColors) {
		if (!modifyOptions)
			return 0; // not enough colors
		if (cNumColors >= 8)
			*nrOfColors = 8;
		else
			*nrOfColors = 2;
		return 1;	// success by reducing nr of colors
	}
	return 1;
}

void Win32Gui::SetColors(struct sGuiOptions *pOptions)
{
	LOGPALETTE *pLog;
	int scale;
	struct sRGBDef  *pc;
	Word    red   = 255;
	Word    green = 255;
	Word    blue  = 255;
	int     i, idx;

	pc = (struct sRGBDef *)&colors;
	while(pc->colorName != NULL) {
		if (strcmp(pOptions->color, pc->colorName) == 0)
			break;
		pc++;
	}
	if (pc->colorName != NULL) {
		red   = pc->red;
		green = pc->green;
		blue  = pc->blue;
	}

	pLog = (LOGPALETTE *)new char[sizeof(LOGPALETTE) + ((1 << COLOR_PLANES) -1) *
				sizeof(PALETTEENTRY)]; 
	pLog->palVersion = 0x300;
	pLog->palNumEntries = 1 << COLOR_PLANES;
	for (i = 0; i < (1 << COLOR_PLANES); i++) {
		idx = i;
		if (pOptions->inverse)
			idx = (1 << COLOR_PLANES) - idx - 1;
		pLog->palPalEntry[idx].peFlags = PC_NOCOLLAPSE;
		if (use_colors) {
			// DEPENDANCIES:
			// the color plane masks used here depend on
			// the same masks used in CopyToZPixmap
			scale  = i & 0x20 ? 2 : 0; scale |= i & 0x04 ? 1 : 0;
			pLog->palPalEntry[idx].peGreen = 255 * sqrt((double)scale / 3);
			scale  = i & 0x10 ? 2 : 0; scale |= i & 0x02 ? 1 : 0;
			pLog->palPalEntry[idx].peRed   = 255 * sqrt((double)scale / 3);
			scale  = i & 0x08 ? 2 : 0; scale |= i & 0x01 ? 1 : 0;
			pLog->palPalEntry[idx].peBlue  = 255 * sqrt((double)scale / 3);
		} else {
			pLog->palPalEntry[idx].peBlue  = blue  * sqrt((double)i / ((1 << COLOR_PLANES)-1));
			pLog->palPalEntry[idx].peRed   = red   * sqrt((double)i / ((1 << COLOR_PLANES)-1));
			pLog->palPalEntry[idx].peGreen = green * sqrt((double)i / ((1 << COLOR_PLANES)-1));
		}
	}
	palette = CreatePalette(pLog);
	delete [] pLog;
}

void Win32Gui::initialize_after_create(HWND w, struct sGuiOptions *pOptions)
{
	HDC             hdc;
	int             bpp;
	struct sRGBDef  *pc;

	oldX          = 0;
	oldY          = 0;
	Word    red   = 255;
	Word    green = 255;
	Word    blue  = 255;
	int     foregroundIdx;
	int     backgroundIdx;
	int     i, j, idx;
	BITMAPINFO *pbmi1, *pbmi6;

	copy_block = new Byte[WINDOWWIDTH * BLOCKHEIGHT *
		MAX_GUIXSIZE * MAX_GUIYSIZE
		/* * 32 / 8*/ ]; // screen depth on Win32 is 8
	pc = (struct sRGBDef *)&colors;
	while(pc->colorName != NULL) {
		if (strcmp(pOptions->color, pc->colorName) == 0)
			break;
		pc++;
	}
	if (pc->colorName != NULL) {
		red   = pc->red;
		green = pc->green;
		blue  = pc->blue;
	}
	if (!pOptions->inverse) {
		foregroundIdx = 1;
		backgroundIdx = 0;
	} else {
		foregroundIdx = 0;
		backgroundIdx = 1;
	}

	hdc = GetDC(w);
	CheckDeviceSupport(hdc, 1, &nColors);
	SetColors(pOptions);
	for (i = 0; i < MAX_GUIXSIZE; i++) {
		for (j = 0; j < MAX_GUIYSIZE; j++) {
			pbmi1 = (BITMAPINFO *)new char[sizeof(BITMAPINFOHEADER) +
				2 * sizeof(RGBQUAD)];
			pbmi6 = (BITMAPINFO *)new char[sizeof(BITMAPINFOHEADER) +
				(1 << COLOR_PLANES) * sizeof(Word)];

			pbmi1->bmiHeader.biSize           = sizeof(BITMAPINFOHEADER);
			pbmi1->bmiHeader.biPlanes         = 1;
			pbmi1->bmiHeader.biWidth          = BLOCKWIDTH * (i+1);
			pbmi1->bmiHeader.biHeight         = -BLOCKHEIGHT * (j+1);
			pbmi1->bmiHeader.biBitCount       = 1;
			pbmi1->bmiHeader.biCompression    = BI_RGB;
			pbmi1->bmiHeader.biSizeImage      = 0;
			pbmi1->bmiHeader.biXPelsPerMeter  = 0;
			pbmi1->bmiHeader.biYPelsPerMeter  = 0;
			pbmi1->bmiHeader.biClrUsed        = 2;
			pbmi1->bmiHeader.biClrImportant   = 2;
			pbmi1->bmiColors[backgroundIdx].rgbBlue   = 0;
			pbmi1->bmiColors[backgroundIdx].rgbGreen  = 0;
			pbmi1->bmiColors[backgroundIdx].rgbRed    = 0;
			pbmi1->bmiColors[backgroundIdx].rgbReserved=0;
			pbmi1->bmiColors[foregroundIdx].rgbBlue   = (BYTE)blue;
			pbmi1->bmiColors[foregroundIdx].rgbGreen  = (BYTE)green;
			pbmi1->bmiColors[foregroundIdx].rgbRed    = (BYTE)red;
			pbmi1->bmiColors[foregroundIdx].rgbReserved=0;

			bpp = COLOR_PLANES; // specify bits per pixel
			pbmi6->bmiHeader.biSize           = sizeof(BITMAPINFOHEADER);
			pbmi6->bmiHeader.biPlanes         = 1;
			pbmi6->bmiHeader.biWidth          = BLOCKWIDTH * (i+1);
			pbmi6->bmiHeader.biHeight         = -BLOCKHEIGHT * (j+1);
			pbmi6->bmiHeader.biBitCount       = 8;
			pbmi6->bmiHeader.biCompression    = BI_RGB;
			pbmi6->bmiHeader.biSizeImage      = 0;
			pbmi6->bmiHeader.biXPelsPerMeter  = 0;
			pbmi6->bmiHeader.biYPelsPerMeter  = 0;
			pbmi6->bmiHeader.biClrUsed        = 1 << bpp;
			pbmi6->bmiHeader.biClrImportant   = 0;

			Word *wp = (Word *)((Byte *)pbmi6 + sizeof(BITMAPINFOHEADER));
			for (idx = 0; idx < (1 << bpp); idx++)
				*(wp++) = idx;

			image1[i][j]  = CreateDIBitmap(hdc, &pbmi1->bmiHeader, 0, NULL,
				pbmi1, DIB_RGB_COLORS);
			image6[i][j]  = CreateDIBitmap(hdc, &pbmi6->bmiHeader, 0, NULL,
				pbmi6, DIB_PAL_COLORS);
			bmi1[i][j] = pbmi1;
			bmi6[i][j] = pbmi6;
		} // for
	} // for
	// initialize Bitmap with all ones
	PatBlt(hdc, 0, 0, BLOCKWIDTH*guiXSize, BLOCKHEIGHT*guiYSize, WHITENESS);
	ReleaseDC(w, hdc);
	// Periodic 20 ms Timer for 50 Hz display update
	idTimer = SetTimer(w, GUI_TIMER_ID, 20, NULL);
}

void Win32Gui::manage_widget(HWND w, struct sGuiOptions *pOptions)
{
	ShowWindow(w, pOptions->nCmdShow);
}  // manage widget


void Win32Gui::initialize_after_open(HWND w, struct sGuiOptions *pOptions)
{
	//ShowCursor(FALSE);  // make cursor invisible
}

int Win32Gui::gui_type(void)
{
	return GUI_WINDOWS;
}

Win32Gui::Win32Gui(
	Mc6809* x_cpu = NULL,
	Scheduler* x_sched = NULL,
	Inout*  x_io = NULL,
	E2video* x_video = NULL,
	struct sGuiOptions *pOptions = NULL) : idTimer(0),
		is_use_undocumented(false),
		AbstractGui(x_cpu, x_sched, x_io, x_video, pOptions)
{
	initialize(pOptions);
}

Win32Gui::~Win32Gui()
{
	int i, j;


	hAccel     = NULL;
	hAccelHwnd = NULL;
	// release all bitmaps
	for (i = 0; i < MAX_GUIXSIZE; i++) {
		for (j = 0; j < MAX_GUIYSIZE; j++) {
			DeleteObject(image1[i][j]);
			DeleteObject(image6[i][j]);
			delete [] bmi1[i][j];
			delete [] bmi6[i][j];
		} // for
	} // for
	if (hFontFixed != NULL)
		DeleteObject(hFontFixed);
	if (palette != NULL)
		DeleteObject(palette);
	if (IsMenu(menubar))
		DestroyMenu(menubar);
	delete [] copy_block;
	wgui = NULL;
}

#endif // #ifdef WIN32
